import sys,sqlite3
CPI   = 6.0   #Cycles per instruction

class Path:
  def __init__(self,index=0,size=0,speed=0,code='',out=''):
    self.index=index
    self.size=size
    self.speed=speed
    self.code=code
    self.out=out
  def addcode(self,code,size,speed,out):
    self.code+='\n  '+code
    self.size+=size
    self.speed+=speed
    if out!='':
      s=''
      for i in (out+","+self.out.replace("=o5=","=o6=").replace("=o4=","=o5=").replace("=o3=","=o4=").replace("=o2=","=o3=").replace("=o1=","=o2=").replace("=o0=","=o1=")).split(","):
        if i.strip() not in s:
          s+=i+","
      self.out=s
      #print(self.code,s)
      #raw_input()
    return self.score()
  def score(self):
    return (self.size*CPI)**2+self.speed**2
  def __gt__(self,x):
    return self.score()>x.score()
  def __lt__(self,x):
    return self.score()<x.score()
  def __ge__(self,x):
    return self.score()>=x.score()
  def __le__(self,x):
    return self.score()<=x.score()
  def __eq__(self,x):
    return self.score()==x.score()
  def __ne__(self,x):
    return self.score()!=x.score()
  def __str__(self):
    return self.code
  def __call__(self,x):
    return self.index
  def copy(self):
    return Path(self.index, self.size, self.speed, self.code, self.out)
class Node:
  def __init__(self,val='',desc=[]):
    self.value=val
    self.desc=desc
    self.parents=[]
    self.children=[]
  def addparent(self,node):
    node.addhild(self)
    return self
  def addchild(self,node):
    if node not in self.children:
      self.children=[node]+self.children
      if self not in node.parents:
        node.parents+=[self]
    return self
  def child(self,n):
    return self.children[n]
  def parent(self,n):
    return self.parent[n]
  def __str__(self):
    s=str(self.value)+'['
    for i in self.children:
      s+=str(i)+","
    return (s.rstrip(",")+"]").replace("[]",'')
  def __eq__(self,n):
    if self.desc==n.desc and self.value==n.value:
      f=True
      for i in range(self.desc[0]):
        f&=(self.child(i)==n.child(i))
      return f
    else:
      return False
# start with an empty path
paths=[Path()]

def numargs(s):
  if s in ["+","-","*","/","^","<<","\~"]:
    return 2
  if s in ["++","--"] or s.endswith("("):
    return 1
  return 0
def iscommutative(s):
  return s in ['+','*','min(','max(']
def astgen(l):
  k=numargs(l[-1])
  n=Node(l[-1],desc=[k])
  l=l[0:-1]
  if len(l)==0 or k==0:
    return [n,l]
  while k>0:
    c=astgen(l)
    n.addchild(c[0])
    l=c[1]
    k-=1
  return [n,l]
def astoptimize(n):
  #Optimize each node
  for i in range(n.desc[0]):
    n.children[i]=astoptimize(n.child(i))

  #For commutative operations, put the input with fewest children on the right, or constant inputs
  for i in range(n.desc[0]-1):
    if iscommutative(n.value):
      if n.child(i).desc[0]<n.child(i+1).desc[0] or "0"<=n.child(i).value[0]<="9":
        a=n.child(i)
        n.children[i]=n.child(i+1)
        n.children[i+1]=a

  #Perform some replacements. ex, 2*x ==> x<<1
  if n.value=="*" and "0"<=n.child(1).value[0]<="9":
    t=int(n.child(1).value)
    if t==0:
      return Node("0",[0])
    elif t==1:
      #Then I need to replace n with child(0)
      return n.child(0)
    elif (t-1)&t==0:
      k=1
      i=2
      while i!=t:
        i+=i
        k+=1
      n.value="<<"
      n.children[1].value=str(k)
      return n
    else:
      return n
  elif n.value=="+" and n.child(1).value=="1":
    n.children=[n.children[0]]
    n.value="++"
    n.desc=[1]
    return n
  else:
    return n
def asttorpn(n):
  l=[n.value]
  for i in n.children[::-1]:
    l=asttorpn(i)+l
  return l
def compile():
  global src,c,paths
  mx=0
  out=Path(size=1<<32,speed=1<<64)
  while len(paths)>1 or out.index==0:
    #Grab the next instruction
    i=paths[0].index
    i=src[i:i+3]
    p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0]+" "+i[1]+" "+i[2],)).fetchall()
    #print(str(paths[0]),paths[0].index,i)
    #raw_input()
    if p!=[]:
      #Now we need to find all of the potential paths
      for i in p:
        t=True
        for j in i[3].replace("=i","=o").split(","):
          t&=(j.strip()[0:-4] in paths[0].out)
        if t:
          pn=paths[0].copy()
          pn.addcode(i[0],i[1],i[2],i[4])
          pn.index+=3
          paths+=[pn]
        
      paths=paths[1:]
    else:
      p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0]+" "+i[1],)).fetchall()
      if p!=[]:
        #Now we need to find all of the potential paths
        for i in p:
          t=True
          for j in i[3].replace("=i","=o").split(","):
            t&=(j.strip()[0:-4] in paths[0].out)
          if t:
            pn=paths[0].copy()
            pn.addcode(i[0],i[1],i[2],i[4])
            pn.index+=2
            paths+=[pn]
        paths=paths[1:]
      elif i[1]=="\\~":
        for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("\\~",)).fetchall():
          j=list(j)
          j[0]=j[0].replace("**","var_"+i[0])
          p+=[j]
          for i in p:
            t=True
            for j in i[3].replace("=i","=o").split(","):
              t&=(j.strip()[0:-4] in paths[0].out)
            if t:
              pn=paths[0].copy()
              pn.addcode(i[0],i[1],i[2],i[4])
              pn.index+=2
              paths+=[pn]
          paths=paths[1:]
      else:
        p=c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",(i[0],)).fetchall()
        if p==[]:
          if "0"<=i[0]<="9":
            p=[]
            for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("constant",)).fetchall():
              j=list(j)
              j[0]=j[0].replace("**",i[0])
              p+=[j]
          else:
            p=[]
            for j in c.execute("SELECT code,size,speed,input,output,requires,state FROM z80 WHERE ir=?",("var",)).fetchall():
              j=list(j)
              j[0]=j[0].replace("**","var_"+i[0])
              p+=[j]
        #Now we need to find all of the potential paths
        for i in p:
          t=True
          for j in i[3].replace("=i","=o").split(","):
            if j.strip()!='':
              t&=(j.strip()[0:-4] in paths[0].out)
          if t:
            pn=paths[0].copy()
            pn.addcode(i[0],i[1],i[2],i[4])
            pn.index+=1
            paths+=[pn]
        paths=paths[1:]
    paths=sorted(paths)
    if mx<len(paths):
      mx=len(paths)
    if paths==[]:
      paths=[Path()]
    if paths[0].index>=len(src)-2:
      if out>paths[0]:
        out=paths[0]
    while (paths[0]>=out) and len(paths)>1:
      paths=paths[1:]
  print("Searched %d paths" % mx)
  return out
fi=''
fo=''
mode=''
asm=False
if len(sys.argv)==1:
  print("%s source [dest]" % sys.argv[0])
  raise SystemExit
for i in sys.argv[1:]:
  if i[0]=='-':
    pass
  else:
    if fi=='':
      fi=i
    else:
      fo=i
if fo=='':
  s=fi.split('.')
  fo=s[0]
  for i in s[1:-1]:
    fo+='.'+i
  fo+='.asm'
print("Generating Z80 code from %s" % fi)
f=open(fi,"r")
code=f.read()
f.close()

# Split the input
src=code.strip().split('\t')

ast=astgen(src)[0]
print(ast)
ast=astoptimize(ast)
ast=asttorpn(ast)
print(ast)
src=ast+['','']

# Load the description database
conn=sqlite3.connect('irToZ80.db')
c=conn.cursor()

cmpld=compile()

#Close the database
conn.close()

f=open(fo,'w')
f.write(cmpld.code)
f.close()
print("Output Z80 code to %s" % fo)
print("%d bytes" % cmpld.size)